from typing import Any

Dataset: Any
dataset: Any
Expression: Any
field: Any
partitioning: Any
FileSystemDataset: Any
ParquetFragmentScanOptions: Any
ParquetFileFormat: Any
ParquetReadOptions: Any
ParquetFileWriteOptions: Any
write_dataset: Any
